#
# A class that supports the solution of football puzzles
#

from itertools import combinations, permutations, product
from collections import defaultdict, namedtuple
from operator import lt, eq, ge
from re import split
from functools import total_ordering

@total_ordering
class Ft_order_ptsgf(object):
  __slots__ = ()

  def __lt__(self, other):
    return self.pts < other.pts if self.pts != other.pts else self.gf < other.gf

  def __eq__(self, other):
    return self.pts == other.pts and self.gf == other.gf

@total_ordering
class Ft_order_ptsgdgf(object):
  __slots__ = ()

  def __lt__(self, other):
    if self.pts != other.pts:
      return self.pts < other.pts
    sgd, ogd = self.gf - self.ga, other.gf - other.ga
    if sgd != ogd:
      return sgd < ogd
    else:
      return self.gf < other.gf

  def __eq__(self, other):
    sgd, ogd = self.gf - self.ga, other.gf - other.ga
    return self.pts == other.pts and sgd == ogd and self.gf == other.gf

def _ug(g, n):
  return (g[0] - n, g[1] - n) if isinstance(g, (tuple, list)) else g - n

def _tg(g):
  return (g[0] <= 0 and g[1] >= 0) if isinstance(g, (tuple, list)) else g == 0

def _xg(g):
  return max(g) if isinstance(g, (tuple, list)) else g

class Football(object):
  '''
  Create a Football class instance for a football puzzle:

    teams:   the names of teams involved; this can be a list or
             tuple of names, comma or space separated names in
             a string, or a string  of single character team
             designators

    outcomes a string giving the list of possible outcomes
             to consider(a subset from 'uldw' standing for
             unplayed, loss, draw and win, default - 'uldw')

    scoring  A dictionary giving the points for wins, draws
             and losses (the latter is assumed to be zero if
             not specified (default {'w':3, 'd':1})

    order    a string which is either "pts/gd/gf" or "pts/gf"
             indicating the way that league positions are
             determined (default: points, goal difference,
             goals for)

    sep      a non alphabetic character that is used between
             two team names when identifying matches between
             teams (defaults to '_'); this can be empty ('')
             for single character team names but not otherwise
  '''
  def __init__(self, teams, outcomes=None, scoring=None, order=None, sep=None):
    if outcomes is None:
      outcomes = 'uldw'
    if scoring is None:
      scoring = dict(w=3, d=1)
    if sep is None:
      sep = '_'

    self.outcomes = tuple(outcomes)
    self.scoring = scoring
    self.sep = sep
    if isinstance(teams, (tuple, list)):
      self.teams = tuple(teams)
    elif ',' in teams or ' ' in teams:
      self.teams = tuple(re.split('[,|\s]+', teams))
    else:
      self.teams = tuple(teams)

    if order is None or order == "pts/gd/gf":
      class Ftable(Ft_order_ptsgdgf, namedtuple('table', 'u l d w p pts gf ga')):
        pass
    elif order == "pts/gf":
      class Ftable(Ft_order_ptsgf, namedtuple('table', 'u l d w p pts gf ga')):
        pass
    else:
      raise NotImplementedError("table order not recognised")
    self.f_table = Ftable(*([0] * 8))

  def split(self, x):
    return x.split(self.sep) if self.sep and len(x) != 2 else tuple(x)

  def score(self, x):
    if isinstance(x, (list, tuple)):
      x = 'w' if x[0] > x[1] else 'd' if x[0] == x[1] else 'l'
    return self.scoring.get(x, 0)

  def swap(self, x):
    if isinstance(x, str):
      if x in self.outcomes:
        return {'l':'w', 'w':'l'}.get(x, x)
      elif (any(x.startswith(t) for t in self.teams) and
            any(x.endswith(t) for t in self.teams)):
        a, b = self.split(x)
        if {a, b} < set(self.teams):
          return self.sep.join((b, a))
      elif x in 'wl':
        x = 'l' if x == 'w' else 'w'
      elif x not in 'ud':
        raise ValueError
    elif isinstance(x, tuple) and len(x) == 2:
      return (x[1], x[0])
    return x

  def cmp_teams(self, x):
    x, y = self.split(x)
    return self.teams.index(x) < self.teams.index(y)

  def cmp_matches(self, x):
    x, y = self.split(x)
    return self.teams.index(x), self.teams.index(y)

  def table(self, games, team):
    '''
    Given a dictionary of results so far, compile a table
    for the specified team; after outcomes have been set
    (unplayed, loss, draw, win) the returned table will
    include these outcomes but goals for (gf) and goals
    against (ga) will be empty; if matches have been
    played, their results will be included. The positions
    of teams in a league can be determines by comparing
    tables accrding to the order parameter set during
    the class initialisation
    '''
    if team not in self.teams:
      raise ValueError('no such team')
    r = dict((x, 0) for x in 'uldw')
    r['p'] = 0
    r['pts'] = 0
    r['gf'] = 0
    r['ga'] = 0
    for fm, v in games.items():
      if fm.startswith(team):
        if isinstance(v, tuple):
          r['gf'] += v[0]
          r['ga'] += v[1]
          v = 'w' if v[0] > v[1] else 'd' if v[0] == v[1] else 'l'
        if v in {None, 'u'}:
          r['u'] += 1
        else:
          r[v] += 1
          r['p'] += 1
          r['pts'] += self.score(v)
    return self.f_table._make(r[x] for x in r)

  def games(self, games, team, opponents=None, **kw):
    '''
    Play games between the <team> and their <opponents>,
    returning a dictionary giving the match outcomes (i.e
    unplayed, loss, draw, win).

    games      is a dictionary of games outcomes or match
               results that have already been determined

    team       the team whose results are being determined

    opponents  a list of opponents against which to play;
               all opponents still to be played will be
               included if none are specified

    outcomes   if given this must be a list giving the
               possible outcomes for each match being
               played e.g. ('w', 'dl', ...) if its
               length is less than the number of matches
               to be played, the default outcomes will
               be used to complete the list
    '''
    if opponents is None:
      opponents = tuple(x for x in self.teams if x != team)
    else:
      opponents = tuple(opponents)
    if not (team in self.teams and all(x in self.teams for x in opponents)):
      raise ValueError('no such team')
    np = []
    for opp in opponents:
      me = self.sep.join((team, opp))
      op = self.sep.join((opp, team))
      if (me in games) != (op in games):
        raise ValueError("Internal Error")
      elif me not in games:
        np.append((me, op))

    if 'outcomes' not in kw:
      outcomes = [self.outcomes]
    else:
      outcomes = kw['outcomes']
      outcomes = [tuple(x) for x in outcomes]
    if len(outcomes) < len(np):
      outcomes += [self.outcomes] * (len(np) - len(outcomes))

    gcopy = games.copy()
    if len(np) > 1:
      for r in product(*outcomes):
        for (me, op), r in zip(np, r):
          gcopy[me] = r
          gcopy[op] = self.swap(r)
        yield gcopy
    elif np:
      me, op = np[0]
      for r in outcomes[0]:
        gcopy[me] = r
        gcopy[op] = self.swap(r)
        yield gcopy
    else:
      yield gcopy

  def _gen_matches(self, tp, gf, ga, s=()):
    if not tp:
      if _tg(gf) and _tg(ga):
        yield s
    else:
      g, *tp = tp
      tf, ta = _xg(gf), _xg(ga)
      if g == 'w':
        for a in range(1, tf + 1):
          for b in range(min(a, ta + 1)):
            yield from self._gen_matches(tp, _ug(gf, a), _ug(ga, b), s + ((a, b),))
      elif g == 'd':
        for a in range(min(tf, ta) + 1):
          yield from self._gen_matches(tp, _ug(gf, a), _ug(ga, a), s + ((a, a),))
      elif g == 'l':
        for a in range(tf + 1):
          for b in range(a + 1, ta + 1):
            yield from self._gen_matches(tp, _ug(gf, a), _ug(ga, b), s + ((a, b),))
      elif g == 'u':
        yield from self._gen_matches(tp, gf, ga, s + ((None),))
      else:
        raise ValueError('Internal Error')

  def matches(self, games, team, gf, ga, fn=lambda x:True):
    '''
    Play matches for team <team> whose maatch outcomes
    are known (uldw etc.) and return possible scores in
    their matches

    games    the dictionary giving the already known
             outcomes (uldw) and match results for
             matches already played

    team     the team whose matches are to be played

    gf, ga   the goals for and goals against the team
             being considered; if these are known they
             can set; if they are not known a tuple
             can be given to set the minimum and the
             maximum values to be considered

    fn      a function that when applied to a potential
            match score (a, b) will return True if it
            is allowed (beware thaat this will not be
            applied to any matches that the teaam has
            already played.
    '''

    ab, tp = [], []
    for fm, v in games.items():
      if fm.startswith(team):
        if isinstance(v, tuple):
          gf = _ug(gf, v[0])
          ga = _ug(ga, v[1])
        elif isinstance(v, str):
          ab.append(fm)
          tp.append(v)
        elif v is not None:
          raise ValueError("Internal Error")

    res = games.copy()
    if not ab:
      if _tg(gf) and _tg(ga):
        yield res
    else:
      for ss in self._gen_matches(tp, gf, ga):
        if all(fn(s) for s in ss):
          for m, s in zip(ab, ss):
            res[m] = s
            res[self.swap(m)] = self.swap(s)
          yield res

  def results(self, games, team=None, **kw):
    '''
    Given a dictionary of match results (i.e.
    scores in matches) return the results of
    <team>'s matches as a list of tuples
    giving (team's goals, opponent's goals);
    when  <team> is not given the results of
    all matches are returned, in which case
    the matches that will be returned are
    those where the first and second teams
    are in the order specified in the teams
    parameter during class initialisation;

    The keyword <all> will include unplayed
    matches as None; otherwise these will
    be omitted.

    The keyword <ordered> will return matches
    with their scores rather than names in
    sorted order

    A list of matches, a list of match scores
    and a tuple giving the number goals for
    and against the team are returned; if all
    matches are returned the latter will not be
    a tuple but the total goals scored.
    '''
    unp = kw.get('all', False)
    sort = kw.get('ordered', False)
    some = team is not None
    gf, ga, op, sc = 0, 0, [], []
    for fm in sorted(games.keys(), key=self.cmp_matches):
      v = games[fm]
      if (some and fm.startswith(team) or
          not some and self.cmp_teams(fm)):
        if isinstance(v, tuple):
          if sort and v[0] > v[1]:
            op.append(self.swap(fm))
            sc.append((v[1], v[0]))
          else:
            op.append(fm)
            sc.append(v)
          gf += v[0]
          ga += v[1]
        elif unp and v is None:
          op.append(fm)
          sc.append(v)

    return tuple(op), tuple(sc), (gf, ga) if some else (gf + ga)

  def arrange(self, mm, dp, mpt=None, tms=None):
    '''
    Given a set of match scores between a number
    of teams and a list of their points earned,
    return a list of game results for each team
    that are compatible with the points for each
    team

      mm    the match scores
      dp    the list of points for the teams
      mpt   the numbe of matches played between teams
    '''
    if tms is None:
      sz = len(dp)
      tms = {i:() for i in range(sz)}
      if mpt is None:
        mpt = sz - 1
      if len(mm) != sz * mpt // 2:
        raise ValueError
    if not mm:
      if tuple(sum(self.score(x) for x in tms[k]) for k in tms) == dp:
        yield tuple(tms[k] for k in tms.keys())
    else:
      m = mm[0]
      nx = tuple(k for k, v in tms.items() if len(v) < mpt)
      for t1, t2 in permutations(nx, 2):
        st1, st2 = tms[t1], tms[t2]
        tms[t1] += (m,)
        tms[t2] += ((m[1], m[0]),)
        yield from self.arrange(mm[1:], dp, mpt, tms)
        tms[t1], tms[t2] = st1, st2
